@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -1828,9 +1828,17 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
  */
 static void connection_count_up(void)
 {
+#ifdef HAVE_SIGPROCMASK
+	sigset_t oldmask;
+#else
+	int	oldmask;
+#endif
+
+	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(CONN_COUNTER_SEM);
 	Req_info->conn_counter++;
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
+	POOL_SETMASK(&oldmask);
 }
 
 /*
@@ -1839,6 +1847,13 @@ static void connection_count_up(void)
  */
 static void connection_count_down(void)
 {
+#ifdef HAVE_SIGPROCMASK
+	sigset_t oldmask;
+#else
+	int	oldmask;
+#endif
+
+	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(CONN_COUNTER_SEM);
 	/*
 	 * Make sure that we do not decrement too much.  If failed to read
@@ -1851,6 +1866,7 @@ static void connection_count_down(void)
 	if (Req_info->conn_counter > 0)
 		Req_info->conn_counter--;
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
+	POOL_SETMASK(&oldmask);
 }
 
 /*
@@ -1951,7 +1967,7 @@ static void init_system_db_connection(void)
 }
 
 /*
- * Initialize my backend status.
+ * Initialize my backend status and master node id.
  * We copy the backend status to private area so that
  * they are not changed while I am alive.
  */
@@ -1967,4 +1983,6 @@ void pool_initialize_private_backend_status(void)
 		/* my_backend_status is referred to by VALID_BACKEND macro. */
 		my_backend_status[i] = &private_backend_status[i];
 	}
+
+	my_master_node_id = REAL_MASTER_NODE_ID;
 }
