@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -179,6 +179,7 @@ static BackendStatusRecord backend_rec;	/* Backend status record */
 static pid_t worker_pid; /* pid of worker process */
 
 BACKEND_STATUS* my_backend_status[MAX_NUM_BACKENDS];		/* Backend status buffer */
+int my_master_node_id;		/* Master node id buffer */
 
 int myargc;
 char **myargv;
@@ -808,10 +809,20 @@ static void daemonize(void)
 	dup2(i, 1);
 	dup2(i, 2);
 
+	/* close syslog connection for daemonizing */
+	if (pool_config->logsyslog) {
+		closelog();
+	}
+
     fdlimit = sysconf(_SC_OPEN_MAX);
     for (i = 3; i < fdlimit; i++)
 		close(i);
 
+	/* reopen syslog connection after daemonizing */
+	if (pool_config->logsyslog) {
+		openlog(pool_config->syslog_ident, LOG_PID|LOG_NDELAY|LOG_NOWAIT, pool_config->syslog_facility);
+	}
+
 	write_pid_file();
 }
 
@@ -1288,12 +1299,18 @@ void degenerate_backend_set(int *node_id_set, int count)
 	pid_t parent = getppid();
 	int i;
 	bool need_signal = false;
+#ifdef HAVE_SIGPROCMASK
+	sigset_t oldmask;
+#else
+	int	oldmask;
+#endif
 
 	if (pool_config->parallel_mode)
 	{
 		return;
 	}
 
+	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(REQUEST_INFO_SEM);
 	Req_info->kind = NODE_DOWN_REQUEST;
 	for (i = 0; i < count; i++)
@@ -1320,6 +1337,7 @@ void degenerate_backend_set(int *node_id_set, int count)
 		kill(parent, SIGUSR1);
 
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
+	POOL_SETMASK(&oldmask);
 }
 
 /* send promote node request using SIGUSR1 */
@@ -1803,13 +1821,13 @@ static void failover(void)
 		{
 			process_info[i].need_to_restart = 1;
 		}
-
-		/*
-		 * Send restart request to worker child.
-		 */
-		kill(worker_pid, SIGUSR1);
 	}
 
+	/*
+	 * Send restart request to worker child.
+	 */
+	kill(worker_pid, SIGUSR1);
+
 	if (Req_info->kind == NODE_UP_REQUEST)
 	{
 		pool_log("failback done. reconnect host %s(%d)",
