@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pcp_child.c,v 1.3 2007/05/17 06:46:02 yamaguti Exp $
- * 
- * pgpool: a language independent connection pool server for PostgreSQL 
+ * $Header$
+ *
+ * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2007	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -32,6 +32,9 @@
 #ifdef HAVE_NETINET_TCP_H
 #include <netinet/tcp.h>
 #endif
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include <signal.h>
 
@@ -50,22 +53,46 @@
 #include "pcp/pcp_stream.h"
 #include "pcp/pcp.h"
 #include "md5.h"
+#include "pool_config.h"
+#include "pool_process_context.h"
+#include "pool_process_reporting.h"
 
 #define MAX_FILE_LINE_LEN    512
 #define MAX_USER_PASSWD_LEN  128
 
 extern void pcp_set_timeout(long sec);
-static int exit_request; /* non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived */
+volatile sig_atomic_t pcp_exit_request; /* non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived */
 
 static RETSIGTYPE die(int sig);
 static PCP_CONNECTION *pcp_do_accept(int unix_fd, int inet_fd);
 static void unset_nonblock(int fd);
 static int user_authenticate(char *buf, char *passwd_file, char *salt, int salt_len);
-static void pool_random_salt(char *md5Salt);
+static RETSIGTYPE wakeup_handler(int sig);
+static RETSIGTYPE reload_config_handler(int sig);
+static RETSIGTYPE restart_handler(int sig);
+static int pool_detach_node(int node_id, bool gracefully);
+static int pool_promote_node(int node_id, bool gracefully);
 
 extern int myargc;
 extern char **myargv;
 
+static volatile sig_atomic_t pcp_got_sighup = 0;
+volatile sig_atomic_t pcp_wakeup_request = 0;
+static volatile sig_atomic_t pcp_restart_request = 0;
+
+#define CHECK_RESTART_REQUEST \
+	do { \
+		if (pcp_restart_request) \
+		{ \
+		  pool_log("pcp child process received restart request"); \
+		  exit(1); \
+		} \
+		else \
+		{ \
+		  pool_initialize_private_backend_status(); \
+		} \
+    } while (0)
+
 void
 pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 {
@@ -75,13 +102,10 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 	char salt[4];
 	int random_salt = 0;
 	int i;
-
 	char tos;
 	int rsize;
 	char *buf = NULL;
 
-	char psbuf[NI_MAXHOST + 128];
-
 	pool_debug("I am PCP %d", getpid());
 
 	/* Identify myself via ps */
@@ -96,26 +120,32 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 	/* set up signal handlers */
 	signal(SIGTERM, die);
 	signal(SIGINT, die);
-	signal(SIGHUP, die);
+	signal(SIGHUP, reload_config_handler);
 	signal(SIGQUIT, die);
 	signal(SIGCHLD, SIG_DFL);
-	signal(SIGUSR1, SIG_DFL);
-	signal(SIGUSR2, SIG_DFL);
+	signal(SIGUSR1, restart_handler);
+	signal(SIGUSR2, wakeup_handler);
 	signal(SIGPIPE, SIG_IGN);
 	signal(SIGALRM, SIG_IGN);
-	
+
+	/* Initialize my backend status */
+	pool_initialize_private_backend_status();
+
+	/* Initialize process context */
+	pool_init_process_context();
+
 	for(;;)
 	{
 		errno = 0;
 
+		CHECK_RESTART_REQUEST;
+
 		if (frontend == NULL)
 		{
 			frontend = pcp_do_accept(unix_fd, inet_fd);
 			if (frontend == NULL)
-			{
-				/* pcp_do_accept() calls pool_error(), so just exit */
-				exit(1);
-			}
+				continue;
+
 			unset_nonblock(frontend->fd);
 		}
 
@@ -174,7 +204,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 		/* is this connection authenticated? if not disconnect immediately*/
 		if ((! authenticated) && (tos != 'R' && tos != 'M'))
 		{
-			pool_debug("pcp_child: connection not authorized");					
+			pool_debug("pcp_child: connection not authorized");
 			free(buf);
 			buf = NULL;
 			pcp_close(frontend);
@@ -187,7 +217,33 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 		pool_debug("pcp_child: received PCP packet type of service '%c'", tos);
 
 		set_ps_display("PCP: processing a request", false);
-			
+
+		if (tos == 'C' || tos == 'd' || tos == 'D' || tos == 'j' ||
+			tos == 'J' || tos == 'O' || tos == 'T')
+		{
+			if (Req_info->switching)
+			{
+				int len;
+				int wsize;
+				char *msg;
+
+				if (Req_info->kind == NODE_UP_REQUEST)
+					msg = "FailbackInProgress";
+				else if (Req_info->kind == NODE_DOWN_REQUEST)
+					msg = "FailoverInProgress";
+				else if (Req_info->kind == PROMOTE_NODE_REQUEST)
+					msg = "PromotionInProgress";
+				else
+					msg = "OperationInProgress";
+
+				len = strlen(msg) + 1;
+				pcp_write(frontend, "e", 1);
+				wsize = htonl(sizeof(int) + len);
+				pcp_write(frontend, &wsize, sizeof(int));
+				pcp_write(frontend, msg, len);
+			}
+		}
+
 		switch (tos)
 		{
 			case 'R':			/* authentication */
@@ -196,7 +252,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 				if (random_salt)
 				{
-					authenticated = user_authenticate(buf, pcp_conf_file, salt, 4);					
+					authenticated = user_authenticate(buf, pcp_conf_file, salt, 4);
 				}
 				if (!random_salt || !authenticated)
 				{
@@ -267,7 +323,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 				char mesg[16];
 
 				snprintf(mesg, sizeof(mesg), "%d", node_count);
-				
+
 				pcp_write(frontend, "l", 1);
 				wsize = htonl(sizeof(code) +
 							  strlen(mesg)+1 +
@@ -288,7 +344,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 			{
 				int node_id;
 				int wsize;
-				
+
 				BackendInfo *bi = NULL;
 
 				node_id = atoi(buf);
@@ -319,7 +375,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					snprintf(port_str, sizeof(port_str), "%d", bi->backend_port);
 					snprintf(status, sizeof(status), "%d", bi->backend_status);
 					snprintf(weight_str, sizeof(weight_str), "%f", bi->backend_weight);
-				
+
 					pcp_write(frontend, "i", 1);
 					wsize = htonl(sizeof(code) +
 								  strlen(bi->backend_hostname)+1 +
@@ -373,7 +429,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					snprintf(mesg+total_port_len, strlen(port)+1, "%s", port);
 					total_port_len += strlen(port)+1;
 				}
-				
+
 				pcp_write(frontend, "n", 1);
 				wsize = htonl(sizeof(code) +
 							  strlen(process_count_str)+1 +
@@ -388,25 +444,24 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
 					exit(1);
 				}
-				
+
 				free(process_list);
 				free(mesg);
 
 				pool_debug("pcp_child: %d process(es) found", process_count);
 				break;
 			}
-				
+
 			case 'P':			/* process info */
 			{
 				int proc_id;
 				int wsize;
-				
-				ProcessInfo *pi = NULL;
+				int num_proc = pool_config->num_init_children;
+				int i;
 
 				proc_id = atoi(buf);
-				pi = pool_get_process_info(proc_id);
 
-				if (pi == NULL)
+				if ((proc_id != 0) && (pool_get_process_info(proc_id) == NULL))
 				{
 					char code[] = "InvalidProcessID";
 
@@ -430,7 +485,16 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					/* Finally, indicate that all data is sent */
 					char fin_code[] = "CommandComplete";
 
-					snprintf(con_info_size, sizeof(con_info_size), "%d", pool_config->max_pool);
+					POOL_REPORT_POOLS *pools = get_pools(&num_proc);
+
+					if (proc_id == 0)
+					{
+						snprintf(con_info_size, sizeof(con_info_size), "%d", num_proc);
+					}
+					else
+					{
+						snprintf(con_info_size, sizeof(con_info_size), "%d", pool_config->max_pool * NUM_BACKENDS);
+					}
 
 					pcp_write(frontend, "p", 1);
 					wsize = htonl(sizeof(arr_code) +
@@ -442,48 +506,67 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					if (pcp_flush(frontend) < 0)
 					{
 						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						free(pools);
 						exit(1);
 					}
 
 					/* Second, send process information for all connection_info */
-
-					for (i = 0; i < pool_config->max_pool; i++)
+					for (i=0; i<num_proc; i++)
 					{
 						char code[] = "ProcessInfo";
+						char proc_pid[16];
 						char proc_start_time[20];
 						char proc_create_time[20];
 						char majorversion[5];
 						char minorversion[5];
 						char pool_counter[16];
+						char backend_id[16];
+						char backend_pid[16];
+						char connected[2];
 
-						snprintf(proc_start_time, sizeof(proc_start_time), "%ld", pi->start_time);
-						snprintf(proc_create_time, sizeof(proc_create_time), "%ld", pi->connection_info[i].create_time);
-						snprintf(majorversion, sizeof(majorversion), "%d", pi->connection_info[i].major);
-						snprintf(minorversion, sizeof(minorversion), "%d", pi->connection_info[i].minor);
-						snprintf(pool_counter, sizeof(pool_counter), "%d", pi->connection_info[i].counter);
+						if (proc_id != 0 && proc_id != pools[i].pool_pid) continue;
+
+						snprintf(proc_pid, sizeof(proc_pid), "%d", pools[i].pool_pid);
+						snprintf(proc_start_time, sizeof(proc_start_time), "%ld", pools[i].start_time);
+						snprintf(proc_create_time, sizeof(proc_create_time), "%ld", pools[i].create_time);
+						snprintf(majorversion, sizeof(majorversion), "%d", pools[i].pool_majorversion);
+						snprintf(minorversion, sizeof(minorversion), "%d", pools[i].pool_minorversion);
+						snprintf(pool_counter, sizeof(pool_counter), "%d", pools[i].pool_counter);
+						snprintf(backend_id, sizeof(backend_pid), "%d", pools[i].backend_id);
+						snprintf(backend_pid, sizeof(backend_pid), "%d", pools[i].pool_backendpid);
+						snprintf(connected, sizeof(connected), "%d", pools[i].pool_connected);
 
 						pcp_write(frontend, "p", 1);
-						wsize = htonl(sizeof(code) +
-									  strlen(pi->connection_info[i].database)+1 +
-									  strlen(pi->connection_info[i].user)+1 +
-									  strlen(proc_start_time)+1 +
-									  strlen(proc_create_time)+1 +
-									  strlen(majorversion)+1 +
-									  strlen(minorversion)+1 +
-									  strlen(pool_counter)+1 +
-									  sizeof(int));
+						wsize = htonl(	sizeof(code) +
+										strlen(proc_pid)+1 +
+										strlen(pools[i].database)+1 +
+										strlen(pools[i].username)+1 +
+										strlen(proc_start_time)+1 +
+										strlen(proc_create_time)+1 +
+										strlen(majorversion)+1 +
+										strlen(minorversion)+1 +
+										strlen(pool_counter)+1 +
+										strlen(backend_id)+1 +
+										strlen(backend_pid)+1 +
+										strlen(connected)+1 +
+										sizeof(int));
 						pcp_write(frontend, &wsize, sizeof(int));
 						pcp_write(frontend, code, sizeof(code));
-						pcp_write(frontend, pi->connection_info[i].database, strlen(pi->connection_info[i].database)+1);
-						pcp_write(frontend, pi->connection_info[i].user, strlen(pi->connection_info[i].user)+1);
+						pcp_write(frontend, proc_pid, strlen(proc_pid)+1);
+						pcp_write(frontend, pools[i].database, strlen(pools[i].database)+1);
+						pcp_write(frontend, pools[i].username, strlen(pools[i].username)+1);
 						pcp_write(frontend, proc_start_time, strlen(proc_start_time)+1);
 						pcp_write(frontend, proc_create_time, strlen(proc_create_time)+1);
 						pcp_write(frontend, majorversion, strlen(majorversion)+1);
 						pcp_write(frontend, minorversion, strlen(minorversion)+1);
 						pcp_write(frontend, pool_counter, strlen(pool_counter)+1);
+						pcp_write(frontend, backend_id, strlen(backend_id)+1);
+						pcp_write(frontend, backend_pid, strlen(backend_pid)+1);
+						pcp_write(frontend, connected, strlen(connected)+1);
 						if (pcp_flush(frontend) < 0)
 						{
 							pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+							free(pools);
 							exit(1);
 						}
 					}
@@ -496,10 +579,12 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					if (pcp_flush(frontend) < 0)
 					{
 						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						free(pools);
 						exit(1);
 					}
 
 					pool_debug("pcp_child: retrieved process information from shared memory");
+					free(pools);
 				}
 				break;
 			}
@@ -507,7 +592,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 			case 'S':			/* SystemDB info */
 			{
 				int wsize;
-				
+
 				SystemDBInfo *si = NULL;
 				si = pool_get_system_db_info();
 
@@ -582,7 +667,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 						int type_list_offset;
 						DistDefInfo *ddi;
 						int i, j;
-					
+
 						for (i = 0; i < si->dist_def_num; i++)
 						{
 							ddi = &si->dist_def_slot[i];
@@ -602,7 +687,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 								pool_error("pcp_child: malloc() failed. reason: %s", strerror(errno));
 								exit(1);
 							}
-						
+
 							col_list_offset = type_list_offset = 0;
 							for (j = 0; j < ddi->col_num; j++)
 							{
@@ -661,16 +746,23 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 				}
 				break;
 			}
-			
+
 			case 'D':			/* detach node */
+			case 'd':			/* detach node gracefully */
 			{
 				int node_id;
 				int wsize;
 				char code[] = "CommandComplete";
-				
+				bool gracefully;
+
+				if (tos == 'D')
+					gracefully = false;
+				else
+					gracefully = true;
+
 				node_id = atoi(buf);
 				pool_debug("pcp_child: detaching Node ID %d", node_id);
-				notice_backend_error(node_id);
+				pool_detach_node(node_id, gracefully);
 
 				pcp_write(frontend, "d", 1);
 				wsize = htonl(sizeof(code) + sizeof(int));
@@ -689,7 +781,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 				int node_id;
 				int wsize;
 				char code[] = "CommandComplete";
-				
+
 				node_id = atoi(buf);
 				pool_debug("pcp_child: attaching Node ID %d", node_id);
 				send_failback_request(node_id);
@@ -705,7 +797,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 				}
 				break;
 			}
-			
+
 			case 'T':
 			{
 				char mode = buf[0];
@@ -733,7 +825,206 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 
 				break;
 			}
-			
+
+			case 'O': /* recovery request */
+			{
+				int node_id;
+				int wsize;
+				char code[] = "CommandComplete";
+				int r;
+
+				node_id = atoi(buf);
+
+				if ((!REPLICATION &&
+					 !(MASTER_SLAVE &&
+					   !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP))) ||
+					(MASTER_SLAVE &&
+					 !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
+					 node_id == PRIMARY_NODE_ID))
+				{
+					int len;
+					char *msg;
+
+					if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP))
+						msg = "primary server cannot be recovered by online recovery.";
+					else
+						msg = "recovery request is accepted only in replication mode or stereaming replication mode. ";
+
+					len = strlen(msg)+1;
+					pcp_write(frontend, "e", 1);
+					wsize = htonl(sizeof(int) + len);
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, msg, len);
+				}
+				else
+				{
+					pool_debug("pcp_child: start online recovery");
+
+					r = start_recovery(node_id);
+					finish_recovery();
+
+					if (r == 0) /* success */
+					{
+						pcp_write(frontend, "c", 1);
+						wsize = htonl(sizeof(code) + sizeof(int));
+						pcp_write(frontend, &wsize, sizeof(int));
+						pcp_write(frontend, code, sizeof(code));
+					}
+					else
+					{
+						int len = strlen("recovery failed") + 1;
+						pcp_write(frontend, "e", 1);
+						wsize = htonl(sizeof(int) + len);
+						pcp_write(frontend, &wsize, sizeof(int));
+						pcp_write(frontend, "recovery failed", len);
+					}
+				}
+				if (pcp_flush(frontend) < 0)
+				{
+					pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+					exit(1);
+				}
+			}
+				break;
+
+			case 'B': /* status request*/
+			{
+				int nrows = 0;
+				POOL_REPORT_CONFIG *status = get_config(&nrows);
+				int len = 0;
+				/* First, send array size of connection_info */
+				char arr_code[] = "ArraySize";
+				char code[] = "ProcessConfig";
+				/* Finally, indicate that all data is sent */
+				char fin_code[] = "CommandComplete";
+
+				pcp_write(frontend, "b", 1);
+				len = htonl(sizeof(arr_code) + sizeof(int) + sizeof(int));
+				pcp_write(frontend, &len, sizeof(int));
+				pcp_write(frontend, arr_code, sizeof(arr_code));
+				len = htonl(nrows);
+				pcp_write(frontend, &len, sizeof(int));
+
+				if (pcp_flush(frontend) < 0)
+				{
+					pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+					exit(1);
+				}
+
+				for (i = 0; i < nrows; i++)
+				{
+					pcp_write(frontend, "b", 1);
+					len = htonl(sizeof(int)
+						+ sizeof(code)
+						+ strlen(status[i].name) + 1
+						+ strlen(status[i].value) + 1
+						+ strlen(status[i].desc) + 1
+					);
+
+					pcp_write(frontend, &len, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+					pcp_write(frontend, status[i].name, strlen(status[i].name)+1);
+					pcp_write(frontend, status[i].value, strlen(status[i].value)+1);
+					pcp_write(frontend, status[i].desc, strlen(status[i].desc)+1);
+				}
+
+				pcp_write(frontend, "b", 1);
+				len = htonl(sizeof(fin_code) + sizeof(int));
+				pcp_write(frontend, &len, sizeof(int));
+				pcp_write(frontend, fin_code, sizeof(fin_code));
+				if (pcp_flush(frontend) < 0)
+				{
+					pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+					exit(1);
+				}
+
+				free(status);
+
+				pool_debug("pcp_child: retrieved status information");
+				break;
+			}
+
+			case 'J':			/* promote node */
+			case 'j':			/* promote node gracefully */
+			{
+				int node_id;
+				int wsize;
+				char code[] = "CommandComplete";
+				bool gracefully;
+
+				if (tos == 'J')
+					gracefully = false;
+				else
+					gracefully = true;
+
+				node_id = atoi(buf);
+				if ( (node_id < 0) || (node_id >= pool_config->backend_desc->num_backends) )
+				{
+					char code[] = "NodeIdOutOfRange";
+					pool_error("pcp_child: node id %d is not valid", node_id);
+					pcp_write(frontend, "e", 1);
+					wsize = htonl(sizeof(code) + sizeof(int));
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+					if (pcp_flush(frontend) < 0)
+					{
+						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						exit(1);
+					}
+					exit(1);
+				}
+				/* promoting node is reserved to Streaming Replication */
+				if (!MASTER_SLAVE || (strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) != 0))
+				{
+					char code[] = "NotInStreamingReplication";
+					pool_error("pcp_child: not in streaming replication mode, can't promote node id %d", node_id);
+					pcp_write(frontend, "e", 1);
+					wsize = htonl(sizeof(code) + sizeof(int));
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+					if (pcp_flush(frontend) < 0)
+					{
+						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						exit(1);
+					}
+					exit(1);
+				}
+
+				if (node_id == PRIMARY_NODE_ID)
+				{
+					char code[] = "NodeIdAlreadyPrimary";
+					pool_error("pcp_child: specified node is already primary node, can't promote node id %d", node_id);
+					pcp_write(frontend, "e", 1);
+					wsize = htonl(sizeof(code) + sizeof(int));
+					pcp_write(frontend, &wsize, sizeof(int));
+					pcp_write(frontend, code, sizeof(code));
+					if (pcp_flush(frontend) < 0)
+					{
+						pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+						exit(1);
+					}
+					exit(1);
+				}
+
+				pool_debug("pcp_child: promoting Node ID %d", node_id);
+				pool_promote_node(node_id, gracefully);
+
+				pcp_write(frontend, "d", 1);
+				wsize = htonl(sizeof(code) + sizeof(int));
+				pcp_write(frontend, &wsize, sizeof(int));
+				pcp_write(frontend, code, sizeof(code));
+				if (pcp_flush(frontend) < 0)
+				{
+					pool_error("pcp_child: pcp_flush() failed. reason: %s", strerror(errno));
+					exit(1);
+				}
+				break;
+			}
+
+			case 'F':
+				pool_debug("pcp_child: stop online recovery");
+				break;
+
 			case 'X':			/* disconnect */
 				pool_debug("pcp_child: client disconnecting. close connection");
 				authenticated = 0;
@@ -759,7 +1050,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 static RETSIGTYPE
 die(int sig)
 {
-	exit_request = 1;
+	pcp_exit_request = 1;
 
 	pool_debug("PCP child receives shutdown request signal %d", sig);
 
@@ -779,6 +1070,18 @@ die(int sig)
 	exit(0);
 }
 
+static RETSIGTYPE
+wakeup_handler(int sig)
+{
+	pcp_wakeup_request = 1;
+}
+
+static RETSIGTYPE
+restart_handler(int sig)
+{
+	pcp_restart_request = 1;
+}
+
 static PCP_CONNECTION *
 pcp_do_accept(int unix_fd, int inet_fd)
 {
@@ -806,7 +1109,7 @@ pcp_do_accept(int unix_fd, int inet_fd)
 	{
 		if (errno == EAGAIN || errno == EINTR)
 			return NULL;
-		
+
 		pool_error("pcp_child: select() failed. reason: %s", strerror(errno));
 		return NULL;
 	}
@@ -836,6 +1139,12 @@ pcp_do_accept(int unix_fd, int inet_fd)
 		return NULL;
 	}
 
+	if (pcp_got_sighup)
+	{
+		pool_get_config(get_config_file_name(), RELOAD_CONFIG);
+		pcp_got_sighup = 0;
+	}
+
 	pool_debug("I am PCP %d accept fd %d", getpid(), afd);
 
 	if (inet)
@@ -876,7 +1185,7 @@ unset_nonblock(int fd)
 {
 	int var;
 
-	/* set fd to none blocking */
+	/* set fd to non-blocking */
 	var = fcntl(fd, F_GETFL, 0);
 	if (var == -1)
 	{
@@ -901,14 +1210,14 @@ user_authenticate(char *buf, char *passwd_file, char *salt, int salt_len)
 	char packet_password[MAX_USER_PASSWD_LEN+1];
 	char encrypt_buf[(MD5_PASSWD_LEN+1)*2];
 	char file_username[MAX_USER_PASSWD_LEN+1];
-	char file_password[MD5_PASSWD_LEN+1];
+	char file_password[MAX_USER_PASSWD_LEN+1];
 	char *index = NULL;
 	static char line[MAX_FILE_LINE_LEN+1];
 	int i, len;
 
 	/* strcpy() should be OK, but use strncpy() to be extra careful */
 	strncpy(packet_username, buf, MAX_USER_PASSWD_LEN);
-	index = memchr(buf, '\0', MAX_USER_PASSWD_LEN) + 1;
+	index = (char *) memchr(buf, '\0', MAX_USER_PASSWD_LEN) + 1;
 	if (index == NULL)
 	{
 		pool_debug("pcp_child: error while reading authentication packet");
@@ -985,18 +1294,104 @@ user_authenticate(char *buf, char *passwd_file, char *salt, int salt_len)
 	return 0;
 }
 
-/*
- *  pool_random_salt
- */
-static void pool_random_salt(char *md5Salt)
+/* SIGHUP handler */
+static RETSIGTYPE reload_config_handler(int sig)
 {
-	long rand = random();
-
-	md5Salt[0] = (rand % 255) + 1;
-	rand = random();
-	md5Salt[1] = (rand % 255) + 1;
-	rand = random();
-	md5Salt[2] = (rand % 255) + 1;
-	rand = random();
-	md5Salt[3] = (rand % 255) + 1;
+	pcp_got_sighup = 1;
+}
+
+/* Dedatch a node */
+static int pool_detach_node(int node_id, bool gracefully)
+{
+	if (!gracefully)
+	{
+		notice_backend_error(node_id);	/* send failover request */
+		return 0;
+	}
+		
+	/*
+	 * Wait until all frontends exit
+	 */
+	*InRecovery = 1;	/* This wiil ensure that new incoming
+						 * connection requests are blocked */
+
+	if (wait_connection_closed())
+	{
+		/* wait timed out */
+		finish_recovery();
+		return -1;
+	}
+
+	/*
+	 * Now all frontends have gone. Let's do failover.
+	 */
+	notice_backend_error(node_id);		/* send failover request */
+
+	/*
+	 * Wait for failover completed.
+	 */
+	pcp_wakeup_request = 0;
+
+	while (!pcp_wakeup_request)
+	{
+		struct timeval t = {1, 0};
+		select(0, NULL, NULL, NULL, &t);
+	}
+	pcp_wakeup_request = 0;
+
+	/*
+	 * Start to accept incoming connections and send SIGUSR2 to pgpool
+	 * parent to distribute SIGUSR2 all pgpool children.
+	 */
+	finish_recovery();
+
+	return 0;
+}
+
+/* Promote a node */
+static int pool_promote_node(int node_id, bool gracefully)
+{
+	if (!gracefully)
+	{
+		promote_backend(node_id);	/* send promote request */
+		return 0;
+	}
+		
+	/*
+	 * Wait until all frontends exit
+	 */
+	*InRecovery = 1;	/* This wiil ensure that new incoming
+						 * connection requests are blocked */
+
+	if (wait_connection_closed())
+	{
+		/* wait timed out */
+		finish_recovery();
+		return -1;
+	}
+
+	/*
+	 * Now all frontends have gone. Let's do failover.
+	 */
+	promote_backend(node_id);		/* send promote request */
+
+	/*
+	 * Wait for failover completed.
+	 */
+	pcp_wakeup_request = 0;
+
+	while (!pcp_wakeup_request)
+	{
+		struct timeval t = {1, 0};
+		select(0, NULL, NULL, NULL, &t);
+	}
+	pcp_wakeup_request = 0;
+
+	/*
+	 * Start to accept incoming connections and send SIGUSR2 to pgpool
+	 * parent to distribute SIGUSR2 all pgpool children.
+	 */
+	finish_recovery();
+
+	return 0;
 }
