@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -98,7 +98,7 @@ void pool_start_query(POOL_QUERY_CONTEXT *query_context, char *query, int len, N
 		query_context->original_query = query;
 		query_context->rewritten_query = NULL;
 		query_context->parse_tree = node;
-		query_context->virtual_master_node_id = REAL_MASTER_NODE_ID;
+		query_context->virtual_master_node_id = my_master_node_id;
 		pool_set_query_in_progress();
 		session_context->query_context = query_context;
 	}
@@ -177,8 +177,8 @@ void pool_setall_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
-		if ((BACKEND_INFO(i)).backend_status == CON_UP ||
-			(BACKEND_INFO((i)).backend_status == CON_CONNECT_WAIT))
+		if (private_backend_status[i] == CON_UP ||
+			(private_backend_status[i] == CON_CONNECT_WAIT))
 			query_context->where_to_send[i] = true;
 	}
 	return;
@@ -273,7 +273,11 @@ int pool_virtual_master_db_node_id(void)
 	{
 		return sc->query_context->virtual_master_node_id;
 	}
-	return REAL_MASTER_NODE_ID;
+
+	/*
+	 * No query context exists. Returns master node id in private buffer.
+	 */
+	return my_master_node_id;
 }
 
 /*
@@ -324,6 +328,27 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	{
 		pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
 	}
+	else if (MASTER_SLAVE && query_context->is_multi_statement)
+	{
+		/*
+		 * If we are in master/slave mode and we have multi stametemt
+		 * query, we should send it to primary server only. Otherwise
+		 * it is possible to send a write query to standby servers
+		 * because we only use the first element of the multi
+		 * statement query and don't care about the rest.  Typical
+		 * situation where we are bugged by this is, "BEGIN;DELETE
+		 * FROM table;END". Note that from pgpool-II 3.1.0
+		 * transactional statements such as "BEGIN" is unconditionaly
+		 * sent to all nodes(see send_to_where() for more details).
+		 * Someday we might be able to understand all part of multi
+		 * statement queries, but until that day we need this band
+		 * aid.
+		 */
+		if (query_context->is_multi_statement)
+		{
+			pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+		}
+	}
 	else if (MASTER_SLAVE)
 	{
 		POOL_DEST dest;
